from typing import List, Optional

from .constants import Urls


class Credentials:
    """Class to handle PowerBI rest API permissions"""

    def __init__(
        self,
        *,
        tenant_id: str,
        client_id: str,
        secret: str,
        scopes: Optional[List[str]] = None,
    ):
        if scopes is None:
            scopes = [Urls.DEFAULT_SCOPE]
        self.tenant_id = tenant_id
        self.client_id = client_id
        self.secret = secret
        self.scopes = scopes



from typing import Optional

from ....utils import SerializedAsset


def _folder_path(
    projects: SerializedAsset,
    project: dict,
    root: Optional[str] = "",
) -> str:
    """Recursive function to compute folder path with list of projects"""
    path = "/" + str(project["name"]) + (root or "")
    if project["parent_id"] is None:
        return path

    parent_project = next(
        parent_project
        for parent_project in projects
        if parent_project["id"] == project["parent_id"]
    )
    return _folder_path(projects, parent_project, path)


def compute_project_path(projects: SerializedAsset) -> SerializedAsset:
    """Compute folder path with parent project name"""
    for project in projects:
        project["folder_path"] = _folder_path(projects, project)
    return projects



from dataclasses import dataclass
from enum import Enum


class CredentialsKey(Enum):
    """Value enum object for the credentials"""

    CLIENT_SECRET = "client_secret"  # noqa: S105
    CLIENT_ID = "client_id"
    HOST = "host"
    GRANT_TYPE = "grant_type"
    API_TOKEN = "api_token"  # noqa: S105


CLIENT_ALLOWED_KEYS = (
    CredentialsKey.HOST.value,
    CredentialsKey.CLIENT_ID.value,
    CredentialsKey.API_TOKEN.value,
)


@dataclass
class SigmaCredentials:
    """Class to handle Sigma rest API permissions"""

    api_token: str
    client_id: str
    host: str

    @classmethod
    def from_secret(cls, secret: dict) -> "SigmaCredentials":
        credentials = {
            k: v for k, v in secret.items() if k in CLIENT_ALLOWED_KEYS
        }
        return cls(**credentials)



from typing import Iterable, Set, Tuple

from .sdk import Dashboard, LookmlModel


def lookml_explore_names(
    lookmls: Iterable[LookmlModel],
) -> Set[Tuple[str, str]]:
    """
    Explores from the lookml models
    Only valid explores are yielded: with all infos
    """
    model_explores = (
        (model, explore)
        for model in lookmls
        for explore in model.explores or []
    )

    return {
        (model.name, explore.name)
        for model, explore in model_explores
        # accept hidden resources
        if model.name and explore.name
    }


def dashboard_explore_names(
    dashboards: Iterable[Dashboard],
) -> Set[Tuple[str, str]]:
    """Explores that appear in dashboards"""
    elements = (
        element
        for dashboard in dashboards
        for element in dashboard.dashboard_elements or []
    )

    return {
        (element.query.model, element.query.view)
        for element in elements
        if element.query and element.query.model and element.query.view
    }


def explore_names_associated_to_dashboards(
    lookmls: Iterable[LookmlModel],
    dashboards: Iterable[Dashboard],
):
    """Retrieve only explores that are associated to a looker dashboard"""
    return lookml_explore_names(lookmls).intersection(
        dashboard_explore_names(dashboards),
    )



from statistics import variance
from time import time
from typing import List

from .retry import MS_IN_SEC, Retry, RetryStrategy


def _within(value: int, min_: int, max_: int) -> bool:
    return value >= min_ and value <= max_


def test_retry_strategy__jitter():
    retry = Retry(count=2, base_ms=1000, jitter_ms=500)
    jitters = [retry.jitter() for _ in range(5)]

    # boundaries
    assert all(_within(j, 250, 750) for j in jitters)
    # randomness
    assert variance(jitters) > 0


def _iterate_base(retry: Retry, count: int) -> List[int]:
    bases: List[int] = []
    for _ in range(3):
        retry.count += 1
        bases.append(retry.base())
    return bases


def test_retry_strategy__base():
    common_args = {"count": 2, "base_ms": 2000, "jitter_ms": 500}
    # default strategy is constant
    retry = Retry(**common_args, strategy=None)
    assert _iterate_base(retry, 3) == [2000, 2000, 2000]
    retry = Retry(**common_args, strategy=RetryStrategy.CONSTANT)
    assert _iterate_base(retry, 3) == [2000, 2000, 2000]
    # linear
    retry = Retry(**common_args, strategy=RetryStrategy.LINEAR)
    assert _iterate_base(retry, 3) == [2000, 4000, 6000]
    # exponential
    retry = Retry(**common_args, strategy=RetryStrategy.EXPONENTIAL)
    assert _iterate_base(retry, 3) == [2000, 4000, 8000]


def test_retry_strategy__check():
    retry = Retry(count=3, base_ms=100, jitter_ms=10)

    before = time()
    assert retry.check()
    assert retry.check()
    assert retry.check()
    after = time()

    assert retry.check() is False
    delta_ms = int((after - before) * MS_IN_SEC)
    assert _within(delta_ms, 315, 345)



import json
import logging
import os
import sys
from datetime import datetime
from typing import Any, List

import pkg_resources

ENCODING = "utf8"
SUMMARY_FILENAME = "summary.json"

logger = logging.getLogger(__name__)


def timestamped_filename(filename: str, ts: int) -> str:
    """format file with a timestamp prefix"""
    return f"{ts}-{filename}"


def get_summary_filename(output_directory: str, ts: int) -> str:
    """return the summary filename"""
    filename = timestamped_filename(SUMMARY_FILENAME, ts)
    return os.path.join(output_directory, filename)


def get_output_filename(name: str, output_directory: str, ts: int) -> str:
    """return the output filename, including directory and timestamp prefix"""
    return os.path.join(output_directory, f"{ts}-{name}.json")


def write_json(filename: str, data: Any):
    """
    write the data to a json file at path filename
    """
    with open(filename, "w", encoding=ENCODING) as f:
        json.dump(data, f)
        logger.info(f"Wrote output file: {filename}")


def _current_version() -> str:
    """fetch the current version of castor extractor running"""
    packages = pkg_resources.require("castor-extractor")
    return str(packages[0].version)


def write_summary(output_directory: str, ts: int, **kwargs):
    """
    write a json file containing extraction session information
    """
    summary = get_output_filename("summary", output_directory, ts)
    write_json(
        summary,
        {
            "python_version": sys.version,
            "version": _current_version(),
            "executed_at": datetime.fromtimestamp(ts).isoformat(),
            **kwargs,
        },
    )


def get_summary_payload(client_info: dict, dt: datetime) -> dict:
    """
    return the payload to be stored in the summary file
    """
    return {
        **client_info,
        "executed_at": dt.isoformat(),
    }


def write_errors_logs(output_directory: str, ts: int, errors: List[str]):
    """
    write a json file logs from code execution
    """
    filename = get_output_filename("errors_logs", output_directory, ts)
    write_json(filename, errors)



from enum import Enum


class LookerAsset(Enum):
    """Looker assets"""

    CONNECTIONS = "connections"
    CONTENT_VIEWS = "content_views"
    DASHBOARDS = "dashboards"
    EXPLORES = "explores"
    FOLDERS = "folders"
    GROUPS_HIERARCHY = "groups_hierarchy"
    GROUPS_ROLES = "groups_roles"
    LOOKML_MODELS = "lookml_models"
    LOOKS = "looks"
    PROJECTS = "projects"
    USERS = "users"
    USERS_ATTRIBUTES = "users_attributes"



from typing import Dict, Iterator, Optional

from ....utils import SerializedAsset
from ..assets import TableauAsset
from ..gql_fields import QUERY_FIELDS
from ..tsc_fields import TSC_FIELDS

QUERY_TEMPLATE = """
{{
  {object_type}Connection(first: {page_size}, after: AFTER_TOKEN_SIGNAL) {{
    nodes {{ {query_fields}
    }}
    pageInfo {{
      hasNextPage
      endCursor
    }}
    totalCount
  }}
}}
"""

RESOURCE_TEMPLATE = "{resource}Connection"


def get_paginated_objects(
    server,
    asset: TableauAsset,
    page_size: int,
) -> SerializedAsset:
    assets: SerializedAsset = []
    for query in QUERY_FIELDS[asset]:
        fields = query["fields"].value
        object_type = query["object_type"].value
        query_formatted = QUERY_TEMPLATE.format(
            object_type=object_type,
            page_size=page_size,
            query_fields=fields,
        )
        resource = RESOURCE_TEMPLATE.format(resource=object_type)
        result_pages = query_scroll(server, query_formatted, resource)
        queried_assets = [asset for page in result_pages for asset in page]
        assets.extend(queried_assets)
    return assets


def query_scroll(
    server,
    query: str,
    resource: str,
) -> Iterator[SerializedAsset]:
    """build a tableau query iterator handling pagination and cursor"""

    def _call(cursor: Optional[str]) -> dict:
        # If cursor is defined it must be quoted else use null token
        token = "null" if cursor is None else f'"{cursor}"'
        query_ = query.replace("AFTER_TOKEN_SIGNAL", token)

        return server.metadata.query(query_)["data"][resource]

    cursor = None
    while True:
        payload = _call(cursor)
        yield payload["nodes"]

        page_info = payload["pageInfo"]
        if page_info["hasNextPage"]:
            cursor = page_info["endCursor"]
        else:
            break


def extract_asset(asset: Dict, asset_type: TableauAsset) -> Dict:
    """Agnostic function extracting dedicated attributes with define asset"""
    return {key: getattr(asset, key) for key in TSC_FIELDS[asset_type]}



"""
This file is a proxy to Looker SDK functions using Castor ApiSettings.
"""

from typing import Optional

from looker_sdk.rtl import (
    auth_session,
    requests_transport,
    serialize,
    transport,
)
from looker_sdk.rtl.api_settings import PApiSettings, SettingsConfig
from looker_sdk.sdk import constants
from looker_sdk.sdk.api40 import methods as methods40
from looker_sdk.sdk.api40.models import (
    Dashboard,
    DashboardElement,
    DBConnection,
    Folder,
    GroupHierarchy,
    GroupSearch,
    Look,
    LookmlModel,
    LookmlModelExplore,
    LookmlModelExploreField,
    LookmlModelExploreFieldset,
    LookmlModelExploreJoins,
    LookmlModelNavExplore,
    Project,
    Query,
    User,
)
from typing_extensions import Protocol

from ..env import timeout_second


class Credentials:
    """ValueObject for the credentials"""

    def __init__(
        self,
        *,
        base_url: str,
        client_id: str,
        client_secret: str,
        timeout: Optional[int] = None,
        **_kwargs,
    ):
        self.base_url = base_url
        self.client_id = client_id
        self.client_secret = client_secret
        self.timeout: int = timeout or timeout_second()

    def to_settings_config(self) -> SettingsConfig:
        return SettingsConfig(
            base_url=self.base_url,
            client_id=self.client_id,
            client_secret=self.client_secret,
            timeout=str(self.timeout),
        )


def has_admin_permissions(sdk_: methods40.Looker40SDK) -> bool:
    """Check if any of the current user's permissions include admin rights"""
    current_user_id = sdk_.me().id
    if not current_user_id:
        return False
    current_user_roles = sdk_.user_roles(current_user_id)
    return any(
        "administer" in role.permission_set.permissions
        for role in current_user_roles
        if role.permission_set and role.permission_set.permissions
    )


def init40(config: Credentials) -> methods40.Looker40SDK:
    """Default dependency configuration"""
    settings = ApiSettings(config=config, sdk_version=constants.sdk_version)
    settings.is_configured()
    transport = requests_transport.RequestsTransport.configure(settings)
    return methods40.Looker40SDK(
        auth=auth_session.AuthSession(
            settings,
            transport,
            serialize.deserialize40,
            "4.0",
        ),
        deserialize=serialize.deserialize40,
        serialize=serialize.serialize40,
        transport=transport,
        api_version="4.0",
    )


class CastorApiSettings(PApiSettings):
    """This is an intermediate class that is meant to be extended (for typing purpose)"""

    def read_config(self) -> SettingsConfig:
        raise NotImplementedError()


class ApiSettings(CastorApiSettings):
    """SDK settings with initialisation using a credential object instead of a path to a .ini file"""

    def __init__(self, config: Credentials, sdk_version: Optional[str] = ""):
        """Configure using a config dict"""
        self.config = config.to_settings_config()
        self.verify_ssl = True
        self.base_url = self.config.get("base_url", "")
        self.timeout = config.timeout
        self.headers = {"Content-Type": "application/json"}
        self.agent_tag = f"{transport.AGENT_PREFIX}"
        if sdk_version:
            self.agent_tag += f" {sdk_version}"

    def read_config(self) -> SettingsConfig:
        """Returns a serialization of the credentials"""
        return self.config



from enum import Enum

WS_HOST_TEMPLATE = "wss://{hostname}/app/{{app_id}}"

DEFAULT_HANDLE = -1

ACCESS_DENIED_ERROR_CODE = 5
APP_SIZE_EXCEEDED_ERROR_CODE = 1015

MEASURES_SESSION_PARAMS = (
    {
        "qInfo": {
            "qType": "MeasureList",
        },
        "qMeasureListDef": {
            "qType": "measure",
            "qData": {
                "title": "/title",
                "tags": "/tags",
                "measure": "/qMeasure",
            },
        },
    },
)


class JsonRpcMethod(Enum):
    """
    Set of JSON-RPC methods used to retrieve the list of measures
    """

    OPEN_DOC = "OpenDoc"
    CREATE_SESSION_OBJECT = "CreateSessionObject"
    GET_LAYOUT = "GetLayout"



import functools
import logging
from datetime import date, timedelta
from typing import Callable, Iterator, List, Optional, Sequence, Tuple

from dateutil.utils import today
from looker_sdk.sdk.api40.models import ContentView, UserAttribute

from ....utils import Pager, PagerLogger, SafeMode, past_date, safe_mode
from ..env import page_size
from ..fields import format_fields
from .constants import (
    CONNECTION_FIELDS,
    CONTENT_VIEWS_FIELDS,
    CONTENT_VIEWS_HISTORY_DAYS,
    DASHBOARD_FIELDS,
    FOLDER_FIELDS,
    GROUPS_HIERARCHY_FIELDS,
    GROUPS_ROLES_FIELDS,
    LOOK_FIELDS,
    LOOKML_FIELDS,
    LOOKML_PROJECT_NAME_BLOCKLIST,
    PROJECT_FIELDS,
    USER_FIELDS,
    USERS_ATTRIBUTES_FIELDS,
)
from .sdk import (
    Credentials,
    Dashboard,
    DBConnection,
    Folder,
    GroupHierarchy,
    GroupSearch,
    Look,
    LookmlModel,
    LookmlModelExplore,
    Project,
    User,
    has_admin_permissions,
    init40,
)

logger = logging.getLogger(__name__)


OnApiCall = Callable[[], None]


def _mondays(history_depth_in_days: int) -> Iterator[date]:
    """
    Fetch all Mondays of the elapsed period
    """
    _MSG = "History depth must be strictly positive"
    assert history_depth_in_days > 0, _MSG
    current_date = past_date(history_depth_in_days)
    end_date = today().date()
    while current_date < end_date:
        if current_date.weekday() == 0:
            yield current_date
        current_date += timedelta(days=1)


class ApiPagerLogger(PagerLogger):
    def __init__(self, on_api_call: Optional[OnApiCall]):
        self._on_api_call = on_api_call

    def on_page(self, page: int, count: int):
        logger.info(f"Fetched page {page} / {count} results")
        self._on_api_call and self._on_api_call()

    def on_success(self, page: int, total: int):
        logger.info(f"All page fetched: {page} pages / {total} results")


class ApiClient:
    """Looker client"""

    def __init__(
        self,
        credentials: Credentials,
        on_api_call: OnApiCall = lambda: None,
        safe_mode: Optional[SafeMode] = None,
    ):
        sdk = init40(credentials)
        if not has_admin_permissions(sdk):
            raise PermissionError("User does not have admin access.")
        else:
            self._sdk = sdk
        self._on_api_call = on_api_call
        self._logger = ApiPagerLogger(on_api_call)
        self.per_page = page_size()
        self._safe_mode = safe_mode

    def folders(self) -> List[Folder]:
        """Lists folders of the given Looker account"""

        def _search(page: int, per_page: int) -> Sequence[Folder]:
            return self._sdk.search_folders(
                fields=format_fields(FOLDER_FIELDS),
                per_page=per_page,
                page=page,
            )

        return Pager(_search, logger=self._logger).all(per_page=self.per_page)

    def dashboards(self) -> List[Dashboard]:
        """Lists dashboards of the given Looker account"""

        def _search(page: int, per_page: int) -> Sequence[Dashboard]:
            return self._sdk.search_dashboards(
                fields=format_fields(DASHBOARD_FIELDS),
                per_page=per_page,
                page=page,
            )

        return Pager(_search, logger=self._logger).all(per_page=self.per_page)

    def _search_looks(self) -> List[Look]:
        """
        fetch looks via `search_looks`
        https://developers.looker.com/api/explorer/4.0/methods/Look/search_looks
        """

        def _search(page: int, per_page: int) -> Sequence[Look]:
            return self._sdk.search_looks(
                fields=format_fields(LOOK_FIELDS),
                per_page=per_page,
                page=page,
            )

        logger.info("Use search_looks endpoint to retrieve Looks")
        return Pager(_search, logger=self._logger).all(per_page=self.per_page)

    def _all_looks(self) -> List[Look]:
        """
        fetch looks via `all_looks`
        https://castor.cloud.looker.com/extensions/marketplace_extension_api_explorer::api-explorer/4.0/methods/Look/all_looks
        """
        logger.info("Use all_looks endpoint to retrieve Looks")

        # No pagination : see https://community.looker.com/looker-api-77/api-paging-limits-14598
        return list(self._sdk.all_looks(fields=format_fields(LOOK_FIELDS)))

    def looks(self, all_looks_endpoint: Optional[bool] = False) -> List[Look]:
        """Lists looks of the given Looker account

        By default, uses the endpoint `search_looks` that allow pagination.
        If `all_looks_endpoint` parameter is set to True, it uses `all_looks` endpoint.
        """
        if all_looks_endpoint:
            return self._all_looks()

        return self._search_looks()

    def users(self) -> List[User]:
        """Lists users of the given Looker account"""

        def _search(page: int, per_page: int) -> Sequence[User]:
            # HACK:
            # We use verified_looker_employee=False (filter out Looker employees)
            # Else api returns an empty list when no parameters are specified
            return self._sdk.search_users(
                fields=format_fields(USER_FIELDS),
                per_page=per_page,
                page=page,
                verified_looker_employee=False,
            )

        return Pager(_search, logger=self._logger).all(per_page=self.per_page)

    def lookml_models(self) -> List[LookmlModel]:
        """Iterates LookML models of the given Looker account"""

        models = self._sdk.all_lookml_models(
            fields=format_fields(LOOKML_FIELDS),
        )

        logger.info("All LookML models fetched")
        self._on_api_call()

        return [
            model
            for model in models
            if model.project_name not in LOOKML_PROJECT_NAME_BLOCKLIST
        ]

    def explores(
        self,
        explore_names=Iterator[Tuple[str, str]],
    ) -> List[LookmlModelExplore]:
        """Iterates explores of the given Looker account for the provided model/explore names"""

        @safe_mode(self._safe_mode)
        def _call(model_name: str, explore_name: str) -> LookmlModelExplore:
            explore = self._sdk.lookml_model_explore(model_name, explore_name)

            logger.info(f"Explore {model_name}/{explore_name} fetched")
            self._on_api_call()
            return explore

        explores = [
            _call(model_name, explore_name)
            for model_name, explore_name in explore_names
        ]
        return list(filter(None, explores))

    def connections(self) -> List[DBConnection]:
        """Lists databases connections of the given Looker account"""

        connections = self._sdk.all_connections(
            fields=format_fields(CONNECTION_FIELDS),
        )

        logger.info("All looker connections fetched")
        self._on_api_call()

        return list(connections)

    def projects(self) -> List[Project]:
        """Lists projects of the given Looker account"""

        projects = self._sdk.all_projects(fields=format_fields(PROJECT_FIELDS))

        logger.info("All looker projects fetched")
        self._on_api_call()

        return list(projects)

    def groups_hierarchy(self) -> List[GroupHierarchy]:
        """Lists groups with hierarchy of the given Looker account"""
        groups_hierarchy = self._sdk.search_groups_with_hierarchy(
            fields=format_fields(GROUPS_HIERARCHY_FIELDS),
        )
        logger.info("All looker groups_hierarchy fetched")
        return list(groups_hierarchy)

    def groups_roles(self) -> List[GroupSearch]:
        """Lists groups with roles of the given Looker account"""
        groups_roles = self._sdk.search_groups_with_roles(
            fields=format_fields(GROUPS_ROLES_FIELDS),
        )
        logger.info("All looker groups_roles fetched")
        return list(groups_roles)

    def content_views(self) -> List[ContentView]:
        """
        List the number of views per {user x week x dashboard|look}
        https://cloud.google.com/looker/docs/reference/looker-api/latest/types/ContentView
        """
        content_views: List[ContentView] = []

        for day in _mondays(history_depth_in_days=CONTENT_VIEWS_HISTORY_DAYS):
            formatted_day = day.strftime("%Y-%m-%d")
            logger.info(f"Fetching content views for week {formatted_day}")

            _fetch = functools.partial(
                self._sdk.search_content_views,
                fields=format_fields(CONTENT_VIEWS_FIELDS),
                start_of_week_date=formatted_day,
                user_id="NOT NULL",
            )
            look_views = list(_fetch(look_id="NOT NULL"))
            dashboard_views = list(_fetch(dashboard_id="NOT NULL"))

            content_views.extend(look_views + dashboard_views)

        logger.info(
            f"All looker content views fetched - {len(content_views)} rows",
        )
        return content_views

    def users_attributes(self) -> List[UserAttribute]:
        """Lists user attributes of the given Looker account"""
        user_attributes = list(
            self._sdk.all_user_attributes(
                fields=format_fields(USERS_ATTRIBUTES_FIELDS),
            ),
        )
        logger.info(
            f"All looker user_attributes fetched - {len(user_attributes)} rows",
        )
        self._on_api_call()

        return user_attributes



from ...utils import uri_encode
from ..abstract import SqlalchemyClient

POSTGRES_URI = "postgresql://{user}:{password}@{host}:{port}/{database}"

DEFAULT_PORT = 5432


class PostgresClient(SqlalchemyClient):
    """Postgres client"""

    @staticmethod
    def name() -> str:
        return "PostgreSQL"

    def _engine_options(self, credentials: dict) -> dict:
        return {"connect_args": {"sslmode": "prefer"}}

    def _build_uri(self, credentials: dict) -> str:
        keys = ("user", "password", "host", "port", "database")
        assert all(k in credentials for k in keys)
        return POSTGRES_URI.format(
            user=credentials["user"],
            password=uri_encode(credentials["password"]),
            host=credentials["host"],
            port=credentials.get("port") or DEFAULT_PORT,
            database=credentials["database"],
        )



from ...utils import SerializedAsset, group_by


def compute_usage_views(usages: SerializedAsset) -> SerializedAsset:
    """Compute usages views with group by workbook_id"""
    grouped = group_by("workbook_id", usages)
    aggregated = [
        {
            "workbook_id": key,
            "view_counts": sum(x["total_views"] for x in value),
        }
        for key, value in grouped.items()
    ]
    return aggregated



import logging
from typing import Dict, Iterable, Iterator, Set

from .column import ColumnChecker
from .enums import Issue

logger = logging.getLogger(__name__)

_SEPARATOR = f"{30 * '-'}\n"

FileTemplate = Dict[str, ColumnChecker]  # column_name, column_checker
IssueCounter = Dict[Issue, int]  # occurrences per type of issue


class FileCheckerResults:
    """
    Results produced by FileCheckerRun.
    Gives the number of valid rows and the number of issues encountered.
    """

    def __init__(self):
        self.total_rows: int = 0
        self.valid_rows: int = 0
        self.counter: IssueCounter = {issue: 0 for issue in Issue}
        self.indices: Set[int] = set()

    def summary(self) -> str:
        """
        builds message containing the main information:
        - total rows in the file
        - status (valid / ERROR)
        - number of invalid rows
        """
        invalid = self.total_rows - self.valid_rows
        msg = f"{self.total_rows} rows -- "
        msg += "valid" if invalid == 0 else f"ERROR ({invalid} invalid rows)"
        return msg

    def invalid_rows(self) -> str:
        """
        builds message containing indices of invalid rows
        """
        rows = list(map(str, sorted([index for index in self.indices])))
        if len(rows) > 10:
            rows = [*rows[:5], "...", *rows[-5:]]
        return "Invalid rows: " + ", ".join([f"#{row}" for row in rows])

    def issues(self) -> str:
        """
        builds message containing information about encountered issues
        """
        issues = [
            f"| {issue.name}: {count}"
            for issue, count in self.counter.items()
            if count > 0
        ]
        return "\n".join(issues)

    def is_valid(self) -> bool:
        return self.valid_rows == self.total_rows


class FileCheckerRun:
    """
    Validates the given file content, using template.
    """

    def __init__(
        self,
        content: Iterable[Dict],
        template: FileTemplate,
        file_name: str,
        verbose: bool = False,
    ):
        self.content = content
        self.template = template
        self.file_name = file_name
        self.verbose = verbose
        self.result = FileCheckerResults()
        self.logger = logging.getLogger(__name__)

    def _record_issue(self, issue: Issue, index: int):
        self.result.counter[issue] += 1
        self.result.indices.add(index + 1)  # easier to understand for end user

    @staticmethod
    def _issue_description(
        name: str,
        checker: ColumnChecker,
        issue: Issue,
    ) -> str:
        """
        Build message describing issue:
        - name of the column
        - name of the issue
        - additional info for the end user (what was expected)
        """
        msg = f"[{name}] - {issue.name}"
        if issue == Issue.WRONG_DATATYPE:
            msg += f" - Expecting {checker.data_type}"
        if issue == Issue.UNAUTHORIZED_VALUE:
            msg += f" - Expecting one of {checker.enum_values}"
        return msg + "\n"

    def _log(self, index: int, row: dict, issue_log: str) -> None:
        """
        Show the given row and its issues.
        Example:
        ```
        Issues detected on Row #3
          id                   3
          name                 christina
          gender               woman
          birth_date           1956-04-01
          description
          height               1.57
          folder_id            9999
        ------------------------------
        [gender]    - UNAUTHORIZED_VALUE - Expecting one of {'FEMALE', 'MALE'}
        [folder_id] - UNKNOWN_REFERENCE
        ------------------------------
        ```
        """
        if not self.verbose:
            return
        header = f"Issues detected on Row #{index+1}\n"
        for k, v in row.items():
            header += f"{str(k):<20} {str(v):<100}\n"
        self.logger.info(header + _SEPARATOR + issue_log + _SEPARATOR)

    def occurrences(self, name: str) -> Set[str]:
        """
        Return values of the given column, provided:
        - the column exists in the template
        - the column is set to UNIQUE (which means occurrences are recorded)
        """
        assert name in self.template, f"Unknown column in template: {name}"
        checker = self.template[name]
        msg = "Cannot return occurrences of non-unique column"
        assert checker.is_unique, msg
        return checker.occurrences

    def summary(self):
        """
        Show general information after running checks:
        - file name & validation status (OK / ERROR)
        - number of rows (invalid/total)
        - summary of the issues: rows index + count per issue type
        """
        msg = f"{self.file_name.upper()} -- {self.result.summary()}\n"
        if not self.result.is_valid():
            msg += self.result.invalid_rows() + "\n"
            msg += self.result.issues() + "\n"
        logger.info(msg + "\n")

    def validate(self) -> None:
        """Validate the whole file without using the content"""
        for _ in self.valid_rows():
            pass

    def valid_rows(self) -> Iterator[Dict]:
        """
        Reads the file content and yields only valid rows.
        - Invalid rows are ignored
        - Show detailed issues and the associated row content if verbose

        After reading the whole content, show summary.

        Also, feeds the FileCheckerResults:
        - number of invalid rows
        - number of issues per type

        Example of usage:
        ```
        checker = FileCheckerRun(content, template)
        for valid in checker.valid_rows():
            process(valid)
        if not checker.is_valid():
            raise AssertionError("File cannot be processed, issues detected")
        ```
        """
        for index, row in enumerate(self.content):
            self.result.total_rows += 1
            has_issue = False
            issue_log = ""
            for name, template in self.template.items():
                value = row.get(name)
                issue = template.check(value)
                if not issue:
                    continue
                has_issue = True
                self._record_issue(issue, index)
                issue_log += self._issue_description(name, template, issue)
            if has_issue:
                self._log(index, row, issue_log)
                continue  # don't return the row
            self.result.valid_rows += 1
            yield row
        self.summary()



import argparse
import logging

from castor_extractor.uploader import FileType  # type: ignore
from castor_extractor.uploader import upload  # type: ignore
from castor_extractor.uploader import upload_manifest  # type: ignore

FILE_TYPES = {FileType.QUALITY, FileType.VIZ, FileType.WAREHOUSE}

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")


def _args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-c",
        "--credentials",
        required=True,
        help="""Path to credentials or credentials as string""",
    )
    parser.add_argument(
        "-s",
        "--source_id",
        required=True,
        help="source id provided by castor",
    )
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("-f", "--file_path", help="path to file to upload")
    group.add_argument(
        "-d",
        "--directory_path",
        help="""directory containing the files to upload.
                WARNING: it will upload all the files included
                in the given repository.""",
    )
    supported_file_type = [ft.value for ft in FileType]
    parser.add_argument(
        "-t",
        "--file_type",
        help="type of file to upload, currently supported are {}".format(
            supported_file_type,
        ),
        choices=supported_file_type,
    )
    parsed = parser.parse_args()
    return {
        "credentials": parsed.credentials,
        "source_id": parsed.source_id,
        "file_path": parsed.file_path,
        "directory_path": parsed.directory_path,
        "file_type": FileType(parsed.file_type),
    }


def main():
    params = _args()

    file_type = params.get("file_type")
    if file_type in FILE_TYPES:
        upload(**params)

    if file_type == FileType.DBT:
        dir_path = params.pop("directory_path")
        assert not dir_path
        del params["file_type"]
        upload_manifest(**params)



import logging
from argparse import ArgumentParser

from castor_extractor.visualization import sigma  # type: ignore

logging.basicConfig(level=logging.INFO, format="%(levelname)s - %(message)s")


def main():
    parser = ArgumentParser()

    parser.add_argument("-H", "--host", help="Sigma host")
    parser.add_argument("-c", "--client-id", help="Sigma client ID")
    parser.add_argument("-a", "--api-token", help="Generated API key")
    parser.add_argument("-o", "--output", help="Directory to write to")

    args = parser.parse_args()
    sigma.extract_all(
        host=args.host,
        client_id=args.client_id,
        api_token=args.api_token,
        output_directory=args.output,
    )



from typing import Callable, Dict, List
from uuid import UUID

from .pager import Pager, PagerLogger, PagerOnId


def _make_callback(elements: List[int]) -> Callable[[int, int], List[int]]:
    def _callback(page: int, per: int) -> List[int]:
        _start = (page - 1) * per
        _end = _start + per
        return elements[_start:_end]

    return _callback


ITEMS = list(range(10))


def test_Pager__all():
    """unit test for Pager#all()"""
    pager = Pager(_make_callback(ITEMS))
    # When no argument provided
    assert pager.all() == ITEMS
    # When per page is less than the number of ITEMS
    assert pager.all(per_page=1) == ITEMS
    # When per page is more than the number of ITEMS
    assert pager.all(per_page=1000) == ITEMS


def test_Pager__iterator__pagination():
    """unit test for Pager#iterator() (pagination)"""
    pager = Pager(_make_callback(ITEMS))

    def nb_of_pages(per_page: int) -> int:
        return len([page for page in pager.iterator(per_page=per_page)])

    assert nb_of_pages(per_page=50) == 1
    assert nb_of_pages(per_page=1) == 10
    assert nb_of_pages(per_page=2) == 5
    assert nb_of_pages(per_page=4) == 3


def test_Pager__iterator__logging():
    """unit test for Pager#iterator() (pagination)"""

    class Logger(PagerLogger):
        def __init__(self):
            self.pages = []
            self.total = 0

        def on_page(self, page: int, count: int):
            self.pages.append((page, count))

        def on_success(self, page: int, total: int):
            self.total = total

    logger = Logger()
    pager = Pager(_make_callback(ITEMS), logger=logger)

    pager.all(per_page=6)

    assert logger.pages == [(1, 6), (2, 4), (3, 0)]
    assert logger.total == 10


ITEMS_WITH_IDS = [
    {"id": "00000000-0000-0000-0000-00000000000" + str(i)} for i in range(1, 10)
]


def _make_callback_with_ids(
    elements: List[Dict[str, str]],
) -> Callable[[UUID, int], List[Dict[str, str]]]:
    def _callback(max_id: UUID, per: int) -> List[Dict[str, str]]:
        """assumes the elements are sorted by id"""
        to_return: List[Dict[str, str]] = []
        for element in elements:
            if element["id"] > str(max_id):
                to_return.append(element)
            if len(to_return) >= per:
                return to_return
        return to_return

    return _callback


def test_pageronid__all():
    """unit test for PagerOnId#all()"""
    pager = PagerOnId(_make_callback_with_ids(ITEMS_WITH_IDS))
    # When no argument provided
    assert pager.all() == ITEMS_WITH_IDS
    # When per page is less than the number of ITEMS
    assert pager.all(per_page=1) == ITEMS_WITH_IDS
    # When per page is more than the number of ITEMS
    assert pager.all(per_page=1000) == ITEMS_WITH_IDS


def test_pageronid__iterator__pagination():
    """unit test for PagerOnId#iterator() (pagination)"""
    pager = PagerOnId(_make_callback_with_ids(ITEMS_WITH_IDS))

    def nb_of_pages(per_page: int) -> int:
        return len([page for page in pager.iterator(per_page=per_page)])

    assert nb_of_pages(per_page=50) == 1
    assert nb_of_pages(per_page=1) == len(ITEMS_WITH_IDS)
    assert nb_of_pages(per_page=2) == 5
    assert nb_of_pages(per_page=4) == 3



import logging
from typing import List

import tableauserverclient as TSC  # type: ignore

from ....utils import SerializedAsset
from ..assets import TableauAsset
from ..constants import PAGE_SIZE, TABLEAU_SERVER_VERSION
from ..usage import compute_usage_views
from .client_utils import extract_asset, get_paginated_objects
from .credentials import CredentialsApi, CredentialsKey, get_value
from .project import compute_project_path
from .safe_mode import safe_mode_fetch_usage

logger = logging.getLogger(__name__)


class ApiClient:
    """
    Connect to Tableau REST API and fetch main assets.
    Superuser credentials are required.
    https://tableau.github.io/server-client-python/docs/
    """

    def __init__(
        self,
        **kwargs,
    ):
        self._credentials = CredentialsApi(
            user=get_value(CredentialsKey.TABLEAU_USER, kwargs, True),
            password=get_value(CredentialsKey.TABLEAU_PASSWORD, kwargs, True),
            token_name=get_value(
                CredentialsKey.TABLEAU_TOKEN_NAME,
                kwargs,
                True,
            ),
            token=get_value(CredentialsKey.TABLEAU_TOKEN, kwargs, True),
            server_url=get_value(CredentialsKey.TABLEAU_SERVER_URL, kwargs),
            site_id=get_value(CredentialsKey.TABLEAU_SITE_ID, kwargs, True),
        )
        self._server = TSC.Server(self._credentials.server_url)
        self._server.add_http_options({"verify": True})
        self._page_size = PAGE_SIZE
        self._server.version = TABLEAU_SERVER_VERSION
        self._safe_mode = bool(kwargs.get("safe_mode"))
        self.errors: List[str] = []

    @staticmethod
    def name() -> str:
        return "Tableau/API"

    def _user_password_login(self) -> None:
        """Login into Tableau using user and password"""
        self._server.auth.sign_in(
            TSC.TableauAuth(
                self._credentials.user,
                self._credentials.password,
                site_id=self._credentials.site_id,
            ),
        )

    def _pat_login(self) -> None:
        """Login into Tableau using personal authentication token"""
        self._server.auth.sign_in(
            TSC.PersonalAccessTokenAuth(
                self._credentials.token_name,
                self._credentials.token,
                site_id=self._credentials.site_id,
            ),
        )

    def login(self) -> None:
        """Login into Tableau"""

        if self._credentials.user and self._credentials.password:
            logger.info("Logging in using user and password authentication")
            return self._user_password_login()

        if self._credentials.token_name and self._credentials.token:
            logger.info("Logging in using token authentication")
            return self._pat_login()

        raise ValueError(
            """Wrong authentication: you should provide either user and password
             or personal access token""",
        )

    def base_url(self) -> str:
        return self._credentials.server_url

    def _fetch_users(self) -> SerializedAsset:
        """Fetches list of User"""
        return [
            extract_asset(user, TableauAsset.USER)
            for user in TSC.Pager(self._server.users)
        ]

    def _fetch_workbooks(self) -> SerializedAsset:
        """Fetches list of Workbooks"""

        return [
            extract_asset(workbook, TableauAsset.WORKBOOK)
            for workbook in TSC.Pager(self._server.workbooks)
        ]

    def _fetch_usages(self, safe_mode: bool) -> SerializedAsset:
        """Fetches list of Usages"""
        if not safe_mode:
            usages = [
                extract_asset(usage, TableauAsset.USAGE)
                for usage in TSC.Pager(self._server.views, usage=True)
            ]

            return compute_usage_views(usages)

        return safe_mode_fetch_usage(self)

    def _fetch_projects(self) -> SerializedAsset:
        """Fetches list of Projects"""
        return compute_project_path(
            [
                extract_asset(project, TableauAsset.PROJECT)
                for project in TSC.Pager(self._server.projects)
            ],
        )

    def _fetch_workbooks_to_datasource(self) -> SerializedAsset:
        """Fetches workbooks to datasource"""

        return self._fetch_paginated_objects(
            TableauAsset.WORKBOOK_TO_DATASOURCE,
        )

    def _fetch_published_datasources(self) -> SerializedAsset:
        """Fetches list of published datasources"""

        return [
            extract_asset(datasource, TableauAsset.PUBLISHED_DATASOURCE)
            for datasource in TSC.Pager(self._server.datasources)
        ]

    def _fetch_datasources(self) -> SerializedAsset:
        """Fetches both embedded and published datasource"""

        return self._fetch_paginated_objects(
            TableauAsset.DATASOURCE,
        )

    def _fetch_fields(self) -> SerializedAsset:
        """Fetches fields"""
        return self._fetch_paginated_objects(
            TableauAsset.FIELD,
        )

    def _fetch_custom_sql_queries(self) -> SerializedAsset:
        """Fetches custom sql queries"""

        return self._fetch_paginated_objects(
            TableauAsset.CUSTOM_SQL_QUERY,
        )

    def _fetch_custom_sql_tables(self) -> SerializedAsset:
        """Fetches custom sql tables"""

        return self._fetch_paginated_objects(
            TableauAsset.CUSTOM_SQL_TABLE,
        )

    def _fetch_paginated_objects(self, asset: TableauAsset) -> SerializedAsset:
        """Fetches paginated objects"""

        return get_paginated_objects(self._server, asset, self._page_size)

    def fetch(self, asset: TableauAsset) -> SerializedAsset:
        """Fetches the given asset"""
        logger.info(f"Fetching {asset.name}")

        if asset == TableauAsset.USER:
            assets = self._fetch_users()

        if asset == TableauAsset.USAGE:
            assets = self._fetch_usages(self._safe_mode)

        if asset == TableauAsset.PROJECT:
            assets = self._fetch_projects()

        if asset == TableauAsset.WORKBOOK:
            assets = self._fetch_workbooks()

        if asset == TableauAsset.PUBLISHED_DATASOURCE:
            assets = self._fetch_published_datasources()

        if asset == TableauAsset.WORKBOOK_TO_DATASOURCE:
            assets = self._fetch_workbooks_to_datasource()

        if asset == TableauAsset.DATASOURCE:
            assets = self._fetch_datasources()

        if asset == TableauAsset.FIELD:
            assets = self._fetch_fields()

        if asset == TableauAsset.CUSTOM_SQL_TABLE:
            assets = self._fetch_custom_sql_tables()

        if asset == TableauAsset.CUSTOM_SQL_QUERY:
            assets = self._fetch_custom_sql_queries()

        logger.info(f"Fetched {asset.name} ({len(assets)} results)")

        return assets
